(*
 *   FeedReader
 *   Copyright (C) 2003 The FeedReader Developement Group
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *)
unit FeedRetrieval;

interface

Uses
  Windows, Classes, SysUtils,
  XDOM_2_3,
  IdHttp, IdComponent, IdAuthentication, IdURI,
  Dialogs,contnrs,
  UHTML2Text, XFunctions, FReaderProperties, Misc, LanguageSupport;

Type
  TWorkerThread = Class( TThread )
  Private
    FOnUpdateUnreadCount: TThreadMethod;
    FOnShowMessage: TThreadMethod;
    FOnReady: TThreadMethod;
    FOnStatusChange: TThreadMethod;
    FOnProgress: TThreadMethod;
    FOnFeedProgress: TThreadMethod;
    FOnNewHeadlinesAvailable: TThreadMethod;
    FOnErrorRetrievingFeed: TThreadMethod;
  Private
    TmpHeadlines: Tobjectlist;
    fRefreshAllFeeds: Boolean;
    fForceUpdate: Boolean;
    FeedToRefreshID: Real;
    FeedURL: String;

    Procedure SyncUpdateUnreadCount( FeedID: Real; NewUnreadMsgs: Integer );
    procedure SyncNewHeadlinesAvailable( FeedID: Real; NewUnreadMsgs: Integer; Const NewsFlash: String );
    Procedure SyncShowErrorMessage( FeedID: Real; Const Msg: String);
    Procedure SyncShowProgress( Min, Max, Cur: Integer; IsFeedProgress: Boolean );
    Procedure SyncErrorRetrievingFeed( FeedID: Real; Const Msg: String );

    Function FeedNeedsUpdate( Index: Integer ): Boolean;
    Function GetHttpStream( FeedID: Real; Const URL, LastModified, UserName, Password: String ): TStream;
    Procedure ParseHttpStream( FeedID: Real; Stream: TStream );
    Procedure ProcessFeed( Const Title: String; Min, Max, Position: Integer;
                           FeedID: Real; Const URL, LastModified, Username, Password: String );
    Procedure ProcessExtSource( Const Title: String; Min, Max, Position: Integer;
                           FeedID: Real; Const URL, LastModified, Username, Password: String );
    Procedure ProcessAllFeeds;
    procedure OnHttpConnected(Sender: TObject);
    Procedure OnHttpWork(Sender: TObject; AWorkMode: TWorkMode; const AWorkCount: Integer);
    Procedure AddNewHeadlinesToWorkingSet( FeedID: Real );
    Procedure SetFeedUpdateTick( FeedID: Real );
  protected
    procedure Execute; override;
  Public
    CurFeedID: Real;
    CurImageURL:string;
    CurFeedTitle: String;
    CurFeedDescription: String;
    CurFeedHomepageURL: String;
    NewUnreadHeadlinesCount: Integer;
    CurNewsFlash: String;
    ErrorMessage: String;
    StatusMessage: String;

    ProgressMin: Integer;
    ProgressMax: Integer;
    ProgressCur: Integer;

    Constructor Create( AllFeeds, ForceUpdate: Boolean; FeedID: Real; Const URL, Title: String );
  Published
    Property OnUpdateUnreadCount: TThreadMethod read FOnUpdateUnreadCount write FOnUpdateUnreadCount;
    Property OnShowMessage: TThreadMethod read FOnShowMessage write FOnShowMessage;
    Property OnStatusChange: TThreadMethod read FOnStatusChange write FOnStatusChange;
    Property OnReady: TThreadMethod read FOnReady write FOnReady;
    Property OnProgress: TThreadMethod read FOnProgress write FOnProgress;
    Property OnFeedProgress: TThreadMethod read FOnFeedProgress write FOnFeedProgress;
    Property OnNewHeadlinesAvailable: TThreadMethod read FOnNewHeadlinesAvailable write FOnNewHeadlinesAvailable;
    Property OnErrorRetrievingFeed: TThreadMethod read FOnErrorRetrievingFeed write FOnErrorRetrievingFeed;
  End;

implementation

Constructor TWorkerThread.Create( AllFeeds, ForceUpdate: Boolean; FeedID: Real; Const URL, Title: String );
Begin
  Inherited Create( True );

  Self.Priority := tpLowest;
  Self.FreeOnTerminate := True;
  tmpheadlines:=tobjectlist.Create;
  Self.fRefreshAllFeeds := AllFeeds;
  Self.fForceUpdate := ForceUpdate;
  Self.FeedToRefreshID := FeedID;
  Self.FeedURL := URL;
  Self.CurFeedTitle := Title;
End;

procedure TWorkerThread.Execute;
Begin
  Try
    SyncShowProgress( 0, 0, 0, True );

    If Length(FeedURL) = 0 Then
      ProcessAllFeeds
     Else
      ProcessFeed( CurFeedTitle, 0, 100, 50, FeedToRefreshID, FeedURL, '', '', '' );

    SyncShowProgress( 0, 0, 0, False );
    SyncShowProgress( 0, 0, 0, True );

    If Assigned( FOnReady ) Then
      Synchronize( FOnReady );

  Except
    On E: Exception Do
      ;
  End;
End;


Function TWorkerThread.FeedNeedsUpdate( Index: Integer ): Boolean;
Var
  UpdatePeriodTime: TDateTime;
  LastUpdateTime: TDateTime;
  Minutes: Integer;
  UpdatePeriod: Integer;
Begin

  If gSubscribedFeeds[Index].updateperiod = DefaultFeedRefreshIdValue Then // use default?
    UpdatePeriod := gProperties.DefaultFeedRefresh
   Else
    UpdatePeriod := gSubscribedFeeds[Index].updateperiod;

  Case UpdatePeriod Of
    1:  Minutes := 5;   // Every 5 minutes
    2:  Minutes := 10;  // Every 10 minutes
    3:  Minutes := 15;  // Every 15 minutes
    4:  Minutes := 20;  // Every 20 minutes
    5:  Minutes := 25;  // Every 25 minutes
    6:  Minutes := 30;  // Every 30 minutes
    7:  Minutes := 35;  // Every 35 minutes
    8:  Minutes := 40;  // Every 40 minutes
    9:  Minutes := 45;  // Every 45 minutes
    10: Minutes := 50;  // Every 50 minutes
    11: Minutes := 55;  // Every 55 minutes
    12: Minutes := 60;  // Every 60 minutes
    13: Minutes := -1;  // Never
    // 14 is default period
    15: Minutes :=  2 * 60;  // Every 2 hours
    16: Minutes :=  3 * 60;  // Every 3 hours
    17: Minutes :=  4 * 60;  // Every 4 hours
    18: Minutes :=  5 * 60;  // Every 5 hours
    19: Minutes :=  6 * 60;  // Every 6 hours
    20: Minutes :=  8 * 60;  // Every 8 hours
    21: Minutes := 12 * 60;  // Every 12 hours
    22: Minutes := 16 * 60;  // Every 16 hours
    23: Minutes := 24 * 60;  // Every 24 hours
   Else
    Minutes := 15; // Every 15 Minutes
  End;

  If (Minutes < 0) And (Not fForceUpdate) Then Begin
    Result := False;
    Exit;
  End;

  If fForceUpdate Then Begin
    Result := True;
    Exit;
  End;

  UpdatePeriodTime := EncodeTime( Minutes Div 60, Minutes Mod 60, 0, 0 );
  LastUpdateTime := gSubscribedFeeds[Index].LastUpdateTime;
  If (LastUpdateTime + UpdatePeriodTime) <= Now Then
    Result := True
   Else
    Result := False;
End;

Procedure TWorkerThread.SetFeedUpdateTick( FeedID: Real );
Var
  i: Integer;
Begin
  For i := Low(gSubscribedFeeds) To High(gSubscribedFeeds) Do
    If FeedID = gSubscribedFeeds[i].feedid Then Begin
      gSubscribedFeeds[i].LastUpdateTime := Now;
      Exit;
    End;
End;

Procedure TWorkerThread.ProcessAllFeeds;
Var
  iSubscribedFeed: Integer;
Begin
  For iSubscribedFeed:= Low(gSubscribedFeeds) To High(gSubscribedFeeds) Do Begin

    Try

      If (Not fRefreshAllFeeds) Then
        If FeedToRefreshID <> gSubscribedFeeds[iSubscribedFeed].feedid Then
          Continue;

      If fRefreshAllFeeds Then
        If gSubscribedFeeds[iSubscribedFeed].UpdateThread <> Nil Then
          Continue;

      If (gSubscribedFeeds[iSubscribedFeed].feedtype <> 'http') and (gSubscribedFeeds[iSubscribedFeed].feedtype <> 'ext') Then
        Continue;

      gSubscribedFeeds[iSubscribedFeed].UpdateThread := Self;
      Try
        If FeedNeedsUpdate( iSubscribedFeed ) Then Begin


          With gSubscribedFeeds[iSubscribedFeed] Do Begin
            if pos('ext://',gSubscribedFeeds[iSubscribedFeed].link)>0 then
            ProcessExtSource( title,
                         Low(gSubscribedFeeds), High(gSubscribedFeeds), iSubscribedFeed,
                         feedid, link, LastModified, username, password )
            else
            ProcessFeed( title,
                         Low(gSubscribedFeeds), High(gSubscribedFeeds), iSubscribedFeed,
                         feedid, link, LastModified, username, password );

            htmlurl := CurFeedHomepageURL;
            description := CurFeedDescription;
            image:=CurImageUrl;
            SetFeedUpdateTick( feedid );
          End;

        End;

      Finally
        gSubscribedFeeds[iSubscribedFeed].UpdateThread := Nil;
      End;

    Except
    End;

  End; // for
End;


Procedure TWorkerThread.ProcessFeed( Const Title: String; Min, Max, Position: Integer;
                                     FeedID: Real; Const URL, LastModified, Username, Password: String );
Var
  HttpStream: TStream;
Begin
  Try
    ErrorMessage := '';
    CurFeedTitle := Title;
    CurFeedHomepageURL := '';
    CurFeedDescription := '';
    CurImageURL:='';
    SyncShowProgress( Min, Max, Position, False );

    TmpHeadlines:=tobjectlist.create;

    HttpStream := GetHttpStream( FeedID, URL, LastModified, Username, Password );

    Try
      If Assigned(HttpStream) Then
        ParseHttpStream( FeedID, HttpStream )
      else
        raise Exception.Create('Couldn''t fetch HTTP-Stream');

    Finally
      FreeAndNil( HttpStream );
    End;

    AddNewHeadlinesToWorkingSet( FeedID );

    SetFeedUpdateTick( FeedID );

  Except
    on E: Exception do
    begin
      SyncShowErrorMessage(FeedID, E.Message);
    end;
  End;
End;

Procedure TWorkerThread.ProcessExtSource( Const Title: String; Min, Max, Position: Integer;
                                     FeedID: Real; Const URL, LastModified, Username, Password: String );
Var
action_update:taction_update;
get_ext_info:tgetinfo;
ol:tobjectlist;
dllname:string;
dllhandle:thandle;
data:ext_info;
Begin
  Try
    ErrorMessage := '';
    CurFeedTitle := Title;
    CurFeedHomepageURL := '';
    CurFeedDescription := '';
    CurImageURL:='';
    SyncShowProgress( Min, Max, Position, False );

    TmpHeadlines:=tobjectlist.Create;
    //this will fetch external dll data source
            dllname:=copy(Url,7,pos('/',copy(Url,7,length(Url)))-1);
            DLLHandle := LoadLibrary(pchar(dllname));
            if DLLHandle <> 0 then
            begin
              @action_update := GetProcAddress(DLLHandle, 'action_update');
              @get_ext_info := GetProcAddress(DLLHandle, 'get_ext_info');

              if assigned(get_ext_info) and (get_ext_info.action_update) and assigned(action_update) then
                tmpheadlines:=action_update(URL, LastModified, Username, Password);

            end;
    AddNewHeadlinesToWorkingSet( FeedID );

    SetFeedUpdateTick( FeedID );

  Except
    on E: Exception do
    begin
      SyncShowErrorMessage(FeedID, E.Message);
    end;
  End;
End;


Procedure TWorkerThread.SyncShowErrorMessage( FeedID: Real; Const Msg: String);
Begin
  Self.ErrorMessage := Msg;
  If Assigned(FOnShowMessage) Then
    Synchronize( FOnShowMessage );
End;


procedure TWorkerThread.OnHttpConnected(Sender: TObject);
begin
  StatusMessage := 'Connected';
  If Assigned(FOnStatusChange) Then
    Synchronize( Self, FOnStatusChange );
end;


Procedure TWorkerThread.OnHttpWork(Sender: TObject; AWorkMode: TWorkMode; const AWorkCount: Integer);
Var
  Position, Max: Integer;
Begin
  StatusMessage := 'Retrieving';
  If Assigned(FOnStatusChange) Then
    Synchronize( Self, FOnStatusChange );

  Max :=  TIdHTTP(sender).Response.ContentLength;
  Position := AWorkCount;
  If Max <= 0 Then Max := Position * 2;
  SyncShowProgress( 0, Max, Position, True );
End;


Function TWorkerThread.GetHttpStream( FeedID: Real; Const URL,LastModified, UserName, Password: String ): TStream;

  Function BuildCompleteURL( Const AUrl: String ): String;
  Var
    i: Integer;
  Begin
    i := Pos( '://', AUrl );
    If i = 0 Then
      Result := 'http://' + AUrl
     Else
      Result := AUrl;
  End;

Var
  http: TIdHTTP;
  i: Integer;
  ExceptionMessage: string;
Begin
  Result := Nil;
  Try
    http := TIdHTTP.Create( Nil );
    http.ReadTimeout := Feedretrieval_Timeout;
    Try
      Result := TMemoryStream.Create;

      http.HandleRedirects := True;
      http.Request.UserAgent := gProperties.UserAgent;
      InitHttpClientProxySettings( http );
      http.OnConnected := OnHttpConnected;
      http.OnWork := OnHttpWork;

      If Length(UserName) > 0 Then Begin
        http.Request.Username := Username;
        http.Request.Password := Password;
        http.Request.BasicAuthentication := True;
      End;
      if LastModified<>'' then begin
        http.Request.CustomHeaders.Add('If-Modified-Since: '+LastModified);
      end;
      ExceptionMessage := '';
      try
        http.Get( BuildCompleteURL( URL ), Result );
      except
        on E: Exception do
        begin
          if (E.Message <> http.ResponseText) then
            ExceptionMessage := BuildCompleteURL( URL ) + #13#10 + E.Message + #13#10
          else
            ExceptionMessage := BuildCompleteURL( URL ) + #13#10 + E.Message;
        end;
      end;
      If (http.ResponseCode <> 200) and (http.ResponseCode <> 304) Then Begin
        SyncErrorRetrievingFeed( FeedID, ExceptionMessage );
        FreeAndNil( Result );
      End;
      For i := Low(gSubscribedFeeds) To High(gSubscribedFeeds) Do
      If FeedID = gSubscribedFeeds[i].feedid Then Begin
        gSubscribedfeeds[i].LastModified:=http.Response.RawHeaders.values['Last-Modified'];
        break;
      End;
    Finally
      http.Free;
    End;
  Except
    On E: Exception Do
      SyncErrorRetrievingFeed( FeedID, E.Message );
  End;
End;


function ExtGetTextContent( Node: TDomNode): wideString;
var
  childType: TdomNodeType;
  childItem: TdomNode;
  i,cl: integer;
begin
  case Node.nodeType of
    ntElement_Node,ntEntity_Reference_Node,ntEntity_Node,ntDocument_Fragment_Node: begin
      result:= '';
      cl:= pred(Node.childnodes.length);
      for i:= 0 to cl do begin
        childItem:= Node.childnodes.item(i);
        childType:= childItem.nodeType;
        if childType = ntEntity_Reference_Node Then result := concat(result, '&' + ChildItem.NodeName + ';')
        else if (childType <> ntComment_Node) and (childType <> ntProcessing_Instruction_Node) then result:= concat(result, ExtGetTextContent( childItem ) );
      end;
    end;
    ntAttribute_Node,ntText_Node,ntCDATA_Section_Node,ntComment_Node,ntProcessing_Instruction_Node:
      result:= Node.nodeValue;
  else
    result:= '';
  end;
end;


Procedure TWorkerThread.ParseHttpStream( FeedID: Real; Stream: TStream );
Var
  domdoc: tdomdocument;
  DomImplementation1: tdomimplementation;
  XmlToDomParser1: txmltodomparser;
  Ch: Char;
  FixedStream: TStream;
  ImageURL: String;
  x2: tdomnodelist;
  z: tdomnode;
  i: Integer;
  TmpNode: TdomNode;
  TmpText,tmplinktitle: String;
  myheadline:theadline;
Begin
  domdoc := Nil;
  DomImplementation1 := Nil;
  XmlToDomParser1 := Nil;
  Try
    DomImplementation1 := tdomimplementation.create(nil);
    XmlToDomParser1 := txmltodomparser.create(nil);
    xmltodomparser1.DOMImpl := domimplementation1;
    domdoc := tdomdocument.create(domimplementation1, Nil );

    TmpHeadlines.clear;

    Stream.Position := 0;
    Try

      Try
        XmlToDomParser1.KeepEntityRefs := False;
        xmltodomparser1.docStreamToDom( Stream, '', '', domdoc );
      Except
        On E: Exception Do
          If E Is EParserException Then Begin
            // sometimes the feed-stream starts with illegal characters,
            // these are removed and we try to parse it again
            Stream.Position := 0;
            While Stream.Position < Stream.Size Do Begin
              Stream.Read( Ch, SizeOf(Ch) );
              If Ch = '<' Then Begin
                Stream.Position := Stream.Position - 1;
                Break;
              End;
            End;
            FixedStream := TMemoryStream.Create;
            Try
              FixedStream.CopyFrom( Stream, Stream.Size-Stream.Position );
              FixedStream.Position := 0;
              xmltodomparser1.docStreamToDom(FixedStream, '', '', domdoc);
            Finally
              FreeAndNil( FixedStream );
            End;
          End
           Else Raise(e);
      End;

    Except
      On e: EConvertError Do Begin
        SyncShowErrorMessage( FeedID, E.Message );
        Exit;
      End;
      On e: EParserException Do Begin
        SyncShowErrorMessage( FeedID, E.Message );
        Exit;
      End;
    End; // Except


    //
    // extract image-url
    //
    ImageURL := '';
    x2 := domdoc.getElementsByTagName('image');
    if x2.length > 0 then
    begin
      if x2.length = 2 then
        z := x2.item(1)
      else
        z := x2.item(0);
      if z.getFirstChildElement('url') <> nil then begin
        z.getFirstChildElement('url').resolveEntityReferences( erReplace );
        ImageURL := z.getFirstChildElement('url').textContent;

      End;
      CurImageURL:=ImageURL;
    end;


    //
    // extract homepage-url
    //
    CurFeedHomepageURL := '';
    x2 := domdoc.getElementsByTagName('channel');
    if x2.length > 0 then
    begin
      z := x2.item(0);
      if z.getFirstChildElement('link') <> nil then begin
        z.getFirstChildElement('link').resolveEntityReferences( erReplace );
        CurFeedHomepageURL := ExtGetTextContent( z.getFirstChildElement('link') );
      End;
    end;

    CurFeedDescription := '';


    //
    // extract homepage-url and description  - ATOM STANDARD
    //

        x2 := domdoc.getElementsByTagName('feed');
    if x2.length > 0 then
    begin
      z := x2.item(0);
      tmpnode:=z.getFirstChildElement('link');
      while tmpnode<> nil do begin
        if tmpnode.hasAttributes and (TmpNode.attributes.getNamedItem('rel')<>nil) and (TmpNode.attributes.getNamedItem('rel').textContent='alternate') and (TmpNode.attributes.getNamedItem('href')<>nil) then
        CurFeedHomepageURL :=         TmpNode.attributes.getNamedItem('href').textContent;
        tmpnode:=tmpnode.getNEXTSIBLINGElement('link');
      End;
      tmpnode:=z.getfirstchildelement('tagline');
        if tmpnode<>nil  then CurFeedDescription :=         TmpNode.textContent;

    end;


    //
    // extract description
    //
    x2 := domdoc.getElementsByTagName('channel');
    if x2.length > 0 then
    begin
      z := x2.item(0);
      if z.getFirstChildElement('description') <> nil then begin
        z.getFirstChildElement('description').resolveEntityReferences( erReplace );
        CurFeedDescription := z.getFirstChildElement('description').textContent;
      End;
    end;


    //
    // extract all items (entry), ATOM standard
    //
    x2 := domdoc.getElementsBytagname('entry');
    for i := 0 to x2.length - 1 do
    begin
      myheadline:=theadline.Create;
      tmpheadlines.add(myheadline);
      z := x2.item(i);
      if z.getFirstChildElement('title') <> nil
        then Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel := stringreplace(stringreplace(html2normal(z.getFirstChildElement('title').textContent), '<title>', '', [rfreplaceall]), '</title>', '', [rfreplaceall]); //MH: textContent????
      if z.getFirstChildElement('content') <> nil
      then begin
        TmpNode := z.getFirstChildElement('content');
        TmpText := ExtGetTextContent( TmpNode );
        if (TmpNode.hasAttributes) and (TmpNode.attributes.getNamedItem('type')<>nil) and  ((TmpNode.attributes.getNamedItem('type').textContent='application/xhtml+xml') or (TmpNode.attributes.getNamedItem('type').textContent='text/html'))
        then Theadline(tmpHeadlines[tmpHeadlines.count - 1]).sisu := ((TmpText)) //MH: textContent????
        else Theadline(tmpHeadlines[tmpHeadlines.count - 1]).sisu := z.getFirstChildElement('content').textContent;
      end;
      if (z.getFirstChildElement('summary') <> nil) and (Theadline(tmpHeadlines[tmpHeadlines.count - 1]).sisu='')
      then begin
        TmpNode := z.getFirstChildElement('summary');
        TmpText := ExtGetTextContent( TmpNode );
        if (TmpNode.hasAttributes) and (TmpNode.attributes.getNamedItem('type')<>nil) and  ((TmpNode.attributes.getNamedItem('type').textContent='application/xhtml+xml') or (TmpNode.attributes.getNamedItem('type').textContent='text/html'))
        then Theadline(tmpHeadlines[tmpHeadlines.count - 1]).sisu := ((TmpText)) //MH: textContent????
        else Theadline(tmpHeadlines[tmpHeadlines.count - 1]).sisu := z.getFirstChildElement('summary').textContent;
      end;
      if z.getFirstChildElement('dc:subject') <> nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).topic := z.getFirstChildElement('dc:subject').textContent;
      if z.getFirstChildElement('id') <> nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).guid := z.getFirstChildElement('id').textContent;
      if z.getFirstChildElement('created') <> nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).date := ConvertRDFDateTimeToInternalFormat(z.getFirstChildElement('created').textContent);

      if z.getFirstChildElement('modified') <> nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).modified := z.getFirstChildElement('modified').textContent;

      if (z.getFirstChildElement('issued') <> nil) and (Theadline(tmpHeadlines[tmpHeadlines.count - 1]).date='') then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).date := ConvertRDFDateTimeToInternalFormat(z.getFirstChildElement('issued').textContent);

      if z.getFirstChildElement('author') <> nil then
        begin
        tmpnode:=z.getFirstChildElement('author');
        if tmpnode.getfirstchildelement('name')<>nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).creator := tmpnode.getFirstChildElement('name').textContent;
        end;

      if z.getFirstChildElement('link') <> nil then
      begin
       TmpNode := z.getFirstChildElement('link');
        while tmpnode<>nil do
        begin
         if (TmpNode.hasAttributes) and (TmpNode.attributes.getNamedItem('rel')<>nil) and ((TmpNode.attributes.getNamedItem('rel').textcontent='alternate')) and (TmpNode.attributes.getNamedItem('href')<>nil) then
         Theadline(tmpHeadlines[tmpHeadlines.count - 1]).link := TmpNode.attributes.getNamedItem('href').textContent;
         if (TmpNode.hasAttributes) and (TmpNode.attributes.getNamedItem('rel')<>nil) and ((TmpNode.attributes.getNamedItem('rel').textcontent='related')) and (TmpNode.attributes.getNamedItem('href')<>nil) then
         begin
              if (TmpNode.attributes.getNamedItem('title')<>nil) then tmplinktitle:=html2normal(TmpNode.attributes.getNamedItem('title').textContent) else tmplinktitle:=TmpNode.attributes.getNamedItem('href').textContent;
              Theadline(tmpHeadlines[tmpHeadlines.count - 1]).other :=          Theadline(tmpHeadlines[tmpHeadlines.count - 1]).other +'<br>'+'Related link=<a href='+TmpNode.attributes.getNamedItem('href').textContent+'>' + tmplinktitle+'</a>';
         end;
         if (TmpNode.hasAttributes) and (TmpNode.attributes.getNamedItem('rel')<>nil) and ((TmpNode.attributes.getNamedItem('rel').textcontent='via')) and (TmpNode.attributes.getNamedItem('href')<>nil) then
         begin
              if (TmpNode.attributes.getNamedItem('title')<>nil) then tmplinktitle:=html2normal(TmpNode.attributes.getNamedItem('title').textContent) else tmplinktitle:=TmpNode.attributes.getNamedItem('href').textContent;
              Theadline(tmpHeadlines[tmpHeadlines.count - 1]).other :=          Theadline(tmpHeadlines[tmpHeadlines.count - 1]).other +'<br>'+'Via=<a href='+TmpNode.attributes.getNamedItem('href').textContent+'>' + tmplinktitle+'</a>';
         end;
         TmpNode :=tmpnode.getNextSiblingElement('link');
//                .getFirstChildElement('link');

        end;
      end;

      if (Theadline(tmpHeadlines[tmpHeadlines.count - 1]).date = '') then Theadline(tmpHeadlines[tmpHeadlines.count - 1]).date := ConvertRFC822ToInternalFormat('');
    end;
    //
    // extract all items, RSS standard
    //
    x2 := domdoc.getElementsBytagname('item');
    for i := 0 to x2.length - 1 do
    begin
      myheadline:=theadline.Create;
      tmpheadlines.add(myheadline);
      z := x2.item(i);
      if ImageURL <> '' then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).image := ImageURL;

            if (z.attributes <> Nil) and (z.attributes.getNamedItem('rdf:about')<>nil) then
              begin
                 Theadline(tmpHeadlines[tmpHeadlines.count - 1]).guid:=z.attributes.getNamedItem('rdf:about').nodeValue;
              end;

      if z.getFirstChildElement('title') <> nil
        then Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel := stringreplace(stringreplace(html2normal(z.getFirstChildElement('title').textContent), '<title>', '', [rfreplaceall]), '</title>', '', [rfreplaceall]); //MH: textContent????

      //if title is empty, lets take title from description field
      if Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel = '' then
      begin
        if z.getFirstChildElement('description')<>nil then
        begin
          z.getFirstChildElement('description').resolveEntityReferences( erReplace );
          Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel := copy(html2text(stringreplace(stringreplace((z.getFirstChildElement('description').textContent), '<description>', '', [rfreplaceall]), '</description>', '', [rfreplaceall])), 1, 120); //MH: textContent????
        end;

        if (z.getFirstChildElement('dc:description')<>nil) and (Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel = '') then
        begin
          z.getFirstChildElement('dc:description').resolveEntityReferences( erReplace );
          Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel := copy(html2text(stringreplace(stringreplace((z.getFirstChildElement('dc:description').textContent), '<dc:description>', '', [rfreplaceall]), '</dc:description>', '', [rfreplaceall])), 1, 120); //MH: textContent????
        end;

        if pos(#13,Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel)>0
            then Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel:=copy(Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel,1,pos(#13,Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel)-1);

        if length(Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel) > 110 then
          Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel :=
            Theadline(tmpHeadlines[tmpHeadlines.count - 1]).tiitel + '...';
      end;
      //end of taking title from description field
      if z.getFirstChildElement('link')<>nil then Begin
        TmpNode := z.getFirstChildElement('link');
        TmpText := ExtGetTextContent( TmpNode );
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).link := TmpText;
      End;

      if z.getFirstChildElement('source')<>nil then
      begin
        TmpNode := z.getFirstChildElement('source');
        TmpText := ExtGetTextContent( TmpNode );
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).SourceDescription := ((TmpText)); //MH: textContent????
        If TmpNode.attributes <> Nil Then Begin
          TmpNode := TmpNode.attributes.getNamedItem('url');
          If TmpNode <> Nil Then
            Theadline(tmpHeadlines[tmpHeadlines.count - 1]).SourceURL := TmpNode.textContent;
        End;
      end;

      if z.getFirstChildElement('description')<>nil then
      begin
        TmpNode := z.getFirstChildElement('description');
        TmpText := ExtGetTextContent( TmpNode );
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).sisu := ((TmpText)); //MH: textContent????
      end;
      if z.getFirstChildElement('content:encoded')<>nil then
      begin
        TmpNode := z.getFirstChildElement('content:encoded');
        TmpText := ExtGetTextContent( TmpNode );
        if length(Theadline(tmpHeadlines[tmpHeadlines.count - 1]).sisu)<length(tmptext) then Theadline(tmpHeadlines[tmpHeadlines.count - 1]).sisu := ((TmpText)); //MH: textContent????
      end;

      if z.getFirstChildElement('pubDate') <> nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).date := ConvertRFC822ToInternalFormat(z.getFirstChildElement('pubDate').textContent);

      if z.getFirstChildElement('guid') <> nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).guid := z.getFirstChildElement('guid').textContent;


      if z.getFirstChildElement('dc:date') <> nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).date := ConvertRDFDateTimeToInternalFormat(z.getFirstChildElement('dc:date').textContent);
      if z.getFirstChildElement('dc:creator') <> nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).creator := z.getFirstChildElement('dc:creator').textContent;
      if z.getFirstChildElement('dc:description')<>nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).sisu := ((z.getFirstChildElement('dc:description').textContent)); //MH: textContent????
      if z.getFirstChildElement('dc:subject') <> nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).topic := z.getFirstChildElement('dc:subject').textContent;
      if z.getFirstChildElement('category') <> nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).topic := z.getFirstChildElement('category').textContent;
      if z.getFirstChildElement('comments') <> nil then
        Theadline(tmpHeadlines[tmpHeadlines.count - 1]).comments := z.getFirstChildElement('comments').textContent;

      //by default enclosure is false
      Theadline(tmpHeadlines[tmpHeadlines.count - 1]).enclosure:=false;

      if z.getFirstChildElement('enclosure') <> nil then
            begin
            TmpNode := z.getFirstChildElement('enclosure');
            TmpText := ExtGetTextContent( TmpNode );
            if (TmpNode.hasAttributes) and (TmpNode.attributes.getNamedItem('url')<>nil) then
              begin
                 Theadline(tmpHeadlines[tmpHeadlines.count - 1]).enclosure:=true;
                 Theadline(tmpHeadlines[tmpHeadlines.count - 1]).enc_url:=TmpNode.attributes.getNamedItem('url').nodeValue;
              end;

            if (TmpNode.hasAttributes) and (TmpNode.attributes.getNamedItem('type')<>nil) then
                 Theadline(tmpHeadlines[tmpHeadlines.count - 1]).enc_type:=TmpNode.attributes.getNamedItem('type').nodeValue;

            if (TmpNode.hasAttributes) and (TmpNode.attributes.getNamedItem('length')<>nil) then
                 Theadline(tmpHeadlines[tmpHeadlines.count - 1]).enc_size:=TmpNode.attributes.getNamedItem('length').nodeValue;

            end;

      if (Theadline(tmpHeadlines[tmpHeadlines.count - 1]).date = '') then Theadline(tmpHeadlines[tmpHeadlines.count - 1]).date := ConvertRFC822ToInternalFormat('');

    end; // for

    // cleanup
    domdoc.clear;
    domImplementation1.clear;

  Finally
    xmltodomparser1.Free;
    domimplementation1.Free;
    domdoc.free;
  End;

End;


Procedure TWorkerThread.AddNewHeadlinesToWorkingSet( FeedID: Real );
Var
  NewsFlash: String;
  NewEntryCounter: Integer;
  i,headlineindex: Integer;
  myheadline1,myheadline:theadline;
  timestamp,parallelowner:real;
  badphrase,uniqueguid,modified,ratingengine:boolean;
  getrating:tgetrating;
Begin
  //check if we have rating engine
  ratingengine:=false;
  if fileexists(gProperties.ApplicationDirectory+'\rating.dll') then
    begin
    DLLHandle := LoadLibrary(pchar(gProperties.ApplicationDirectory+'\rating.dll'));
    if DLLHandle <> 0 then @getrating := GetProcAddress(DLLHandle, 'getrating');
    if assigned(getrating) then ratingengine:=true;
    end;
  //end of checking ratingengine
  //check if all guids in feed are unique
  try
    uniqueguid:=true;
    for i :=1 to  TmpHeadlines.count-1 do
      if theadline(tmpheadlines[i]).guid=theadline(tmpheadlines[i-1]).guid then
      begin
      uniqueguid:=false;
      break;
      end;
  except
  end;
  Try
    timestamp:=strtofloat(floattostr(TimeStampToMSecs(DateTimeToTimeStamp(now))));
    NewsFlash := '';
    NewEntryCounter := 0;
    for i := TmpHeadlines.count-1 downto 0 do
    begin
      badphrase:=isbadphrase(FeedID,Trim(theadline(TmpHeadlines[i]).tiitel),theadline(TmpHeadlines[i]).sisu);
      if badphrase then Continue;
      headlineindex:=-1;
      if uniqueguid and (Length(theadline(TmpHeadlines[i]).guid) > 0)
          then headlineindex:=GetHeadLineIndex(false,floattostr(feedid)+':'+theadline(TmpHeadlines[i]).guid);

      if headlineindex=-1
          then headlineindex:=GetHeadlineIndex(false,FeedID, UnEscapeHTML(Trim(theadline(TmpHeadlines[i]).tiitel)), theadline(TmpHeadlines[i]).link, -1);

      if (headlineindex>-1) then
          begin
            //found article in database
            if (trim(theadline(TmpHeadlines[i]).modified)<>'')
              and (theadline(gHeadlines[ headlineindex ]).modified=theadline(TmpHeadlines[i]).modified)
                then modified:=false else modified:=true;

            if modified AND ((theadline(gHeadlines[ headlineindex ]).tiitel <> UnEscapeHTML( Trim(theadline(TmpHeadlines[i]).tiitel) ))
               or (theadline(gHeadlines[ headlineindex ]).sisu <> theadline(TmpHeadlines[i]).sisu)
               or (theadline(gHeadlines[ headlineindex ]).link <> theadline(TmpHeadlines[i]).link))
               then
                 begin
                   theadline(gHeadlines[ headlineindex ]).tiitel := UnEscapeHTML( Trim(theadline(TmpHeadlines[i]).tiitel) );
                   theadline(gHeadlines[ headlineindex ]).sisu := theadline(TmpHeadlines[i]).sisu;
                   theadline(gHeadlines[ headlineindex ]).link := theadline(TmpHeadlines[i]).link;
                   theadline(gHeadlines[ headlineindex ]).modified := theadline(TmpHeadlines[i]).modified;
                   if theadline(gHeadlines[ headlineindex ]).read=1 then theadline(gHeadlines[ headlineindex ]).read:=2;
                 end;
            end
      else
            //no article found in database
            begin
            Inc( NewEntryCounter );
            myheadline:=theadline.Create;

           timestamp:=timestamp+1;
            NewsFlash := UnEscapeHTML( Trim(theadline(TmpHeadlines[i]).tiitel) ) + #13 + #10 + NewsFlash;
            myheadline.hlid      := timestamp;
            myheadline.owner     :=  FeedID;
            myheadline.tiitel    := UnEscapeHTML( Trim(theadline(TmpHeadlines[i]).tiitel) );
            myheadline.link      := theadline(TmpHeadlines[i]).link;
            myheadline.SourceDescription := theadline(TmpHeadlines[i]).SourceDescription;
            myheadline.SourceURL := theadline(TmpHeadlines[i]).SourceURL;
            myheadline.linkdoc   := theadline(TmpHeadlines[i]).linkdoc;
            myheadline.sisu      := theadline(TmpHeadlines[i]).sisu;
            myheadline.image     := theadline(TmpHeadlines[i]).image;
            myheadline.modified  := theadline(TmpHeadlines[i]).modified;
            myheadline.topic     := theadline(TmpHeadlines[i]).topic;
            myheadline.creator   := theadline(TmpHeadlines[i]).creator;
            myheadline.enclosure   := theadline(TmpHeadlines[i]).enclosure;
            myheadline.enc_url   := theadline(TmpHeadlines[i]).enc_url;
            myheadline.enc_size   := theadline(TmpHeadlines[i]).enc_size;
            myheadline.enc_type   := theadline(TmpHeadlines[i]).enc_type;

            myheadline.deleted   := false;
            myheadline.comments  := theadline(TmpHeadlines[i]).comments;
            myheadline.section   := theadline(TmpHeadlines[i]).section;
            myheadline.other     := theadline(TmpHeadlines[i]).other;
            myheadline.read    := 0;
            //set rating
            if ratingengine then
              begin
              myheadline.rating    := -1;
              EnterCriticalSection(gCritSect);
              try
              myheadline.rating:=getrating(myheadline);
              except
              end;
              LeaveCriticalSection(gCritSect);
              end;
            //end of seting rating

            myheadline.flag       :=getflag(FeedID,Trim(theadline(TmpHeadlines[i]).tiitel),theadline(TmpHeadlines[i]).sisu);

              If uniqueguid and (Length(theadline(TmpHeadlines[i]).guid) > 0) Then
                myheadline.guid := floattostr(feedid)+':'+theadline(TmpHeadlines[i]).guid
              Else
                myheadline.guid := floattostr(TimeStampToMSecs(DateTimeToTimeStamp(now)));
            myheadline.date := theadline(TmpHeadlines[i]).date;
            misc.gHeadlines.Add(myheadline);
            parallelowner:=get_copy_to(feedid,Trim(theadline(TmpHeadlines[i]).tiitel),theadline(TmpHeadlines[i]).sisu);
            if parallelowner>-1
                then
                  begin
                  myheadline1:=theadline.Create;
                  myheadline1.tiitel:=myheadline.tiitel;
                  myheadline1.link:=myheadline.link;
                  myheadline1.linkdoc:=myheadline.linkdoc;
                  myheadline1.sisu:=myheadline.sisu;
                  myheadline1.image:=myheadline.image;
                  myheadline1.topic:=myheadline.topic;
                  myheadline1.creator:=myheadline.creator;
                  myheadline1.date:=myheadline.date;
                  myheadline1.comments:=myheadline.comments;
                  myheadline1.section:=myheadline.section;
                  myheadline1.other:=myheadline.other;
                  myheadline1.modified:=myheadline.modified;
                  myheadline1.SourceDescription:=myheadline.SourceDescription;
                  myheadline1.SourceURL:=myheadline.SourceURL;
                  myheadline1.deleted:=myheadline.deleted;
                  myheadline1.read:=myheadline.read;
                  myheadline1.flag:=myheadline.flag;
                  myheadline1.rating:=myheadline.rating;
                  myheadline1.hlid:=timestamp+60000;
                  myheadline1.enclosure:=myheadline.enclosure;
                  myheadline1.enc_url:=myheadline.enc_url;
                  myheadline1.enc_size:=myheadline.enc_size;
                  myheadline1.enc_type:=myheadline.enc_type;
              If uniqueguid and (Length(theadline(TmpHeadlines[i]).guid) > 0) Then
                myheadline1.guid := floattostr(parallelowner)+':'+theadline(TmpHeadlines[i]).guid
              Else
                myheadline1.guid := floattostr(TimeStampToMSecs(DateTimeToTimeStamp(now)));
                  myheadline1.itemtype:=myheadline.itemtype;
                  myheadline1.owner:=parallelowner;
                  misc.gHeadlines.Add(myheadline1);
                  end;
            end;

    end; //for
    tmpheadlines.Free;
    Misc.SortHeadlines(gHeadlines);

    SyncUpdateUnreadCount( FeedID, NewEntryCounter );

    If fRefreshAllFeeds And (NewEntryCounter > 0) Then
      SyncNewHeadlinesAvailable( FeedID, NewEntryCounter, NewsFlash );

  Except
    On E: Exception Do
      SyncShowErrorMessage( FeedID, E.Message );
  End;
End;


procedure TWorkerThread.SyncNewHeadlinesAvailable( FeedID: Real; NewUnreadMsgs: Integer; Const NewsFlash: String );
Begin
  If Not Assigned(FOnNewHeadlinesAvailable) Then
    Exit;

  Self.CurFeedID := FeedID;
  Self.NewUnreadHeadlinesCount := NewUnreadMsgs;
  Self.CurNewsFlash := NewsFlash;
  Synchronize( FOnNewHeadlinesAvailable );
End;

Procedure TWorkerThread.SyncUpdateUnreadCount( FeedID: Real; NewUnreadMsgs: Integer );
Begin
  If Not Assigned(FOnUpdateUnreadCount) Then
    Exit;

  Self.CurFeedID := FeedID;
  Self.NewUnreadHeadlinesCount := NewUnreadMsgs;
  Synchronize( FOnUpdateUnreadCount );
End;


Procedure TWorkerThread.SyncShowProgress( Min, Max, Cur: Integer; IsFeedProgress: Boolean );
Begin
  ProgressMin := Min;
  ProgressMax := Max;
  ProgressCur := Cur;

  If IsFeedProgress Then Begin
    If Assigned( FOnFeedProgress ) Then
      Synchronize( FOnFeedProgress );
  End
   Else Begin
     If Assigned( FOnProgress ) Then
       Synchronize( FOnProgress );
   End;
End;


Procedure TWorkerThread.SyncErrorRetrievingFeed( FeedID: Real; Const Msg: String );
Begin
  CurFeedID := FeedID;
  ErrorMessage := Msg;
  If Assigned(FOnErrorRetrievingFeed) Then
    Synchronize( FOnErrorRetrievingFeed );
End;

end.
