unit Formatted_XDOM;

// Version 2.1
//
// The contents of this Package are subject to the Mozilla Public License
// Version 1.1 (the "License"); you may not use this file except in compliance
// with the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
// specific language governing rights and limitations under the License.
//
// The original code is Formatted XML, first released on October 8, 2002.
//
// The initial developer of the original code is Carmi Grushko (sf_yourshadow@bezeqint.net)
//
//----------------------------------------------------------------------------------------------------------------------
//
// History :
// => 08/10/2002 - First release for XDOM 2.3.30+
// => 15/10/2002 - More fixes to make output look more like
//                 MSXML's output.
// => 01/02/2003 - modified to compile with XDOM 2.3.48 and do what FeedReader wants.
//
//                 Tested to work under XDOM 2.3.31.
//
//                 New property: AttributesInNewLine.
//                 Determines if each attribute should be
//                 printed in a new line.
//
// => 16/03/2004 - Modified so that no empty lines are in the output (khoffrath).
//
//----------------------------------------------------------------------------------------------------------------------
//
// Use of these components is fairly easy - use a TPrettyDomToXmlParser
// component to write out XML trees.
//
// => NewLine is the string (CRLF / CR / LF) that will be used to seperate
//    lines in the formatted output.
// => Indent is the number of white spaces each indentation level defines -
//    the children of the document element will be indented 1*Indent, their
//    children 2*Indent and so on.
// => Formatted is a general switch that controls whether the formatting
//    happens from the first place.

interface

uses XDOM_2_3, Classes;

type
  TPrettyXmlStreamBuilder = class(TXmlStreamBuilder)
  private
    nIndentLevel: Integer;
    FFormatted : boolean;
    FIndent : integer;
    FNewLine : TdomNewLineType;
    FAttrNewLine : boolean;
  protected
    function InsertFormatting(const Sender : TXmlCustomReader;
      const Locator: TDomLocator; const Indent: integer;
      NewLine: boolean) : boolean; virtual;
  public
    constructor Create(AOwner : TComponent); override;
    function endElement(const sender: TXmlCustomReader;
                        const locator: TdomLocator;
                              tagName: wideString): boolean; override;
    function startElement(const sender: TXmlCustomReader;
                          const locator: TdomLocator;
                                tagName: wideString;
                                attributes: TdomNameValueList): boolean; override;
    function PCDATA(const sender: TXmlCustomReader;
                    const locator: TdomLocator;
                          data: wideString): boolean; override;
  published
    property Formatted : boolean read FFormatted write FFormatted default true;
    property Indent : integer read FIndent write FIndent default 3;
    property NewLine : TdomNewLineType read FNewLine write FNewLine default nltCRLF;
    property AttributesInNewLine : boolean read FAttrNewLine write FAttrNewLine default false;
  end;

  TPrettyDomToXmlParser = class(TDomToXmlParser)
  private
    function GetAttrNewLine: boolean;
    procedure SetAttrNewLine(const Value: boolean);
  protected
    function GetFormatted : boolean; virtual;
    function GetIndent : integer; virtual;
    function GetNewLine : TdomNewLineType; virtual;
    procedure SetFormatted( const Value : boolean ); virtual;
    procedure SetIndent( const Value : integer ); virtual;
    procedure SetNewLine( const Value : TdomNewLineType ); override;
  public
    constructor Create(AOwner : TComponent); override;
  published
    property Formatted : boolean read GetFormatted write SetFormatted default false;
    property Indent : integer read GetIndent write SetIndent default 3;
    property NewLine : TdomNewLineType read GetNewLine write SetNewLine;
    property AttributesInNewLine : boolean read GetAttrNewLine write SetAttrNewLine;
  end;

procedure Register;

implementation

procedure Register;
begin
     RegisterComponents( 'XML', [TPrettyXmlStreamBuilder, TPrettyDomToXmlParser] );
end;

{ TPrettyXmlStreamBuilder }

constructor TPrettyXmlStreamBuilder.Create(AOwner: TComponent);
begin
     inherited;
     FIndent := 3;
     FFormatted := true;
     FAttrNewLine := false;

     nIndentLevel := 0;
end;

function TPrettyXmlStreamBuilder.endElement(const sender: TXmlCustomReader;
                        const locator: TdomLocator;
                              tagName: wideString): boolean;
begin

     Result := true;
     { Modified by Carmi Grushko to support Formatted Output }
     if Locator.RelatedNode.hasChildNodes then
     begin
          if assigned(FOnEndElement) then FOnEndElement(sender,locator,tagName);

          if FFormatted and (nIndentLevel < 3) then
          begin
            Dec(nIndentLevel);
            result := InsertFormatting( Sender, Locator, FIndent, True );
            Inc(nIndentLevel);
          end;

          if Result then
            result:= writeWideStrings(sender,locator,['</', tagName, '>']);

//###
//          if FFormatted and (nIndentLevel < 1) and
//             not ( (locator.relatedNode.childNodes.length=1) and (locator.relatedNode.firstChild is TdomText) ) then
//            result := InsertFormatting( Sender, Locator, FIndent, True );
     end;
     { End of block }

     Dec(nIndentLevel);
     
     if result
       then if assigned(nextHandler)
         then result:= nextHandler.endElement(sender,locator,tagName);
end;

function TPrettyXmlStreamBuilder.InsertFormatting( const Sender: TXmlCustomReader;
  const Locator: TDomLocator; const Indent: integer; NewLine: boolean): boolean;

  function Spaces : WideString;
  var
    i: integer;
  begin
       Result := '';
       for i := 1 to Indent do
         Result := Result + WideChar(' ');
  end;

const
  NewLineConvert : array[TdomNewLineType] of WideString = (
    ( #13#10 ),
    ( #13 ),
    ( #10 ) );
Var
  i: Integer;
begin
     Result := true;

     if (nIndentLevel > 0) and NewLine then
       Result := WriteWideStrings( Sender, Locator, [ NewLineConvert[ FNewLine ] ] );

     For i := 1 To (nIndentLevel) Do
       Result:= WriteWideStrings( Sender, Locator, [Spaces] );
end;

function TPrettyXmlStreamBuilder.PCDATA(const sender: TXmlCustomReader; const locator: TdomLocator; data: wideString): boolean;
var
  i: integer;
  content: TdomCustomStr;
  modifiedData: wideString;
begin
  if assigned(FOnPCDATA) then FOnPCDATA(sender,locator,data);

  content:= TdomCustomStr.create;
  try
    for i:= 1 to length(data) do begin
      case ord(data[i]) of
        38: content.addWideString('&amp;'); // Ampersand ('&')
        60: content.addWideString('&lt;');  // Less than ('<')
        62: content.addWideString('&gt;');  // Greater than ('>')
        13: content.addWideString('&#xD;'); // Carriage Return (CR)
      else
        content.addWideChar(data[i]);
      end;
    end;
    modifiedData:= content.value;
  finally
    content.free;
  end;

  { Added by Carmi Grushko to support Formatted Output }
  Result := true;
  if FFormatted and (locator <> Nil) And (locator.relatedNode <> Nil) And (locator.relatedNode.parentNode <> Nil) And (locator.relatedNode.parentNode.childNodes <> Nil) And
    (locator.relatedNode.parentNode.childNodes.length > 1)
  then
    ;//Result := InsertFormatting( sender, locator, FIndent, False );
  { End of block }

  if Result then
    result:= writeWideStringPCDATA(sender,locator,modifiedData);

  if result
    then if assigned(nextHandler)
      then result:= nextHandler.PCDATA(sender,locator,data);
end;

function TPrettyXmlStreamBuilder.startElement(const sender: TXmlCustomReader;
                          const locator: TdomLocator;
                                tagName: wideString;
                                attributes: TdomNameValueList): boolean;
var
  i : integer;
  fEmitNewLine: Boolean;
begin

  if assigned(FOnStartElement) then FOnStartElement(sender,locator,tagName,attributes);

  { Added by Carmi Grushko to support Formatted Output }
  Result := true;
  if FFormatted then Begin
    If nIndentLevel > 0 Then Begin
      fEmitNewLine := True;
      If (locator.relatedNode.previousSibling <> Nil) Then
        fEmitNewLine := False;
      Result := InsertFormatting(sender,locator,FIndent, fEmitNewLine );
    End;
  End;
  { End of block }

  Inc( nIndentLevel );

  if Result then
    result:= writeWideStrings(sender,locator,['<', tagName]);

  for i:= 0 to pred(attributes.length) do
  begin
       if not result then
         break;

       if FFormatted and FAttrNewLine then
         result := InsertFormatting(sender,locator,FIndent, False );
       if not result then
         break;

       result:= writeWideStrings(sender,locator,[' ',attributes.names[i],'="']);
       if not result then
         break;

       result:= writeWideStringPCDATA(sender,locator,attributes.values[i]);
       if not result then
         break;

       result:= writeWideString(sender,locator,'"');
  end;

  { Modified by Carmi Grushko to support empty elements }
//  if result
//    then result:= writeWideString(sender,locator,'>');
  if result then
    if locator.relatedNode.hasChildNodes then
      result:= writeWideString(sender,locator,'>')
    else
      result:= writeWideString(sender,locator,'/>');
  { End of block }

  if result
    then if assigned(nextHandler)
      then result:= nextHandler.startElement(sender,locator,tagName,attributes);
end;

constructor TPrettyDomToXmlParser.Create(AOwner: TComponent);
begin
     inherited;
     FStreamBuilder.Free;
     FStreamBuilder:= TPrettyXmlStreamBuilder.create(self);
     FDomReader.NextHandler:= FStreamBuilder;
end;

function TPrettyDomToXmlParser.GetAttrNewLine: boolean;
begin
     Result:= TPrettyXmlStreamBuilder(FStreamBuilder).FAttrNewLine;
end;

function TPrettyDomToXmlParser.getFormatted: boolean;
begin
     Result:= TPrettyXmlStreamBuilder(FStreamBuilder).FFormatted;
end;

function TPrettyDomToXmlParser.getIndent: integer;
begin
     Result:= TPrettyXmlStreamBuilder(FStreamBuilder).FIndent;
end;

function TPrettyDomToXmlParser.GetNewLine: TdomNewLineType;
begin
     Result := TPrettyXmlStreamBuilder(FStreamBuilder).FNewLine;
end;

procedure TPrettyDomToXmlParser.SetAttrNewLine(const Value: boolean);
begin
     TPrettyXmlStreamBuilder(FStreamBuilder).FAttrNewLine := value;
end;

procedure TPrettyDomToXmlParser.setFormatted(const value: boolean);
begin
     TPrettyXmlStreamBuilder(FStreamBuilder).FFormatted:= value;
end;

procedure TPrettyDomToXmlParser.setIndent(const value: integer);
begin
     TPrettyXmlStreamBuilder(FStreamBuilder).FIndent:= value;
end;


procedure TPrettyDomToXmlParser.SetNewLine(const Value: TdomNewLineType);
begin
     TPrettyXmlStreamBuilder(FStreamBuilder).FNewLine := Value;
end;

end.
