(*
 *   FeedReader
 *   Copyright (C) 2003 The FeedReader Developement Group
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *)
unit WebFavorites;

interface

Uses
  Classes, Forms, ComCtrls, SyncObjs;

Type
  TWebFavoriteType = (WFT_Favorite, WFT_Folder);

  TWebFavorite = Class( TObject )
  Public
    FavType: TWebFavoriteType;

    Description: String;
    URL: String;
    FileName: String;

    LastModified: TDateTime;
    Expires: TDateTime;
    LastChecked: TDateTime;
    UpdateInterval: Integer;
  End;

  TWebFavoritesList = Class( TObject )
  Protected
    Iterating: Boolean;
    RootDirectory: String;
    List: TStringList;
    CritSect: TCriticalSection;

    Function BuildIndexValue( Const Filename: String; Const Long: Boolean = False ): String;
    Function IndexValueToFullFileName( Const Index: String ): String;

    Procedure ClearPrim;
    Procedure AddPrim( WebFav: TWebFavorite );
    Procedure DeletePrim( WebFav: TWebFavorite ); overload;
    Procedure DeletePrim( Const Filename: String ); overload;
    Function ItemPrim( Const Filename: String ): TWebFavorite; overload;
    Function ItemPrim( Index: Integer ): TWebFavorite; overload;
    Function CountPrim: Integer;
  Public
    Constructor Create( Const ARootDirectory: String ); virtual;
    Destructor Destroy; override;

    Procedure Clear;

    Procedure Add( WebFav: TWebFavorite );
    Procedure Delete( WebFav: TWebFavorite ); overload;
    Procedure Delete( Const Filename: String ); overload;

    Function Item( Const Filename: String ): TWebFavorite; overload;
    Function Item( Index: Integer ): TWebFavorite; overload;
    Function Count: Integer;

    Procedure BeginIteration;
    Procedure EndIteration;

    Procedure PopulateFavoritesToTree( TreeView: TTreeView; ARootNode: TTreeNode;
                                       FavoriteImageIndex, FolderImageIndex: Integer  );

    Procedure CheckAllFavorites( ARootNode: TTreeNode );

    Procedure SaveToFile( OwnerForm: TForm );
    Procedure LoadFromFile( OwnerForm: TForm );
  End;

implementation

Uses
  Windows, shlobj, SysUtils, ActiveX,
  IEConst, XDOM_2_3, Misc, FReaderProperties,
  IdGlobal, IdHttp, IdComponent, IdAuthentication, IdURI, IdAuthenticationManager;

Function GetFavoritesDirectory: String;
var
  pIdl: pItemIDList;
  Path: PAnsiChar;
begin
  SHGetSpecialFolderLocation(0, CSIDL_FAVORITES, pidl);
  Path := StrAlloc(MAX_PATH);
  if SHGetPathFromIDList(PIDL, Path) then
    Result := path
  Else
   Result := ExtractFilePath( ParamStr(0) );
  StrDispose(Path);

  Result := IncludeTrailingPathDelimiter( Result );
End;


function ResolveUrlIntShCut(Filename: string): string;
var
  IURL: IUniformResourceLocator;
  PersistFile: IPersistfile;
  FName: array[0..MAX_PATH] of WideChar;
  p: Pchar;
begin
  if Succeeded(CoCreateInstance(CLSID_InternetShortcut, nil, CLSCTX_INPROC_SERVER,
    IID_IUniformResourceLocator, IURL))
    then begin
    Persistfile := IUrl as IPersistFile;
    StringToWideChar(FileName, FName, MAX_PATH);
    PersistFile.Load(Fname, STGM_READ);
    IUrl.geturl(@P);
    Result := P;
  end;
end;


Constructor TWebFavoritesList.Create( Const ARootDirectory: String );
Begin
  Inherited Create;

  If Length( ARootDirectory ) = 0 Then
    RootDirectory := IncludeTrailingPathDelimiter( GetFavoritesDirectory )
   Else
    RootDirectory := IncludeTrailingPathDelimiter( ARootDirectory );

  Iterating := False;
  CritSect := TCriticalSection.Create;
  List := TStringList.Create;
  List.Duplicates := dupIgnore;
End;


Destructor TWebFavoritesList.Destroy;
Begin
  Clear;

  List.Free;
  CritSect.Free;

  Inherited Destroy;
End;


Procedure TWebFavoritesList.ClearPrim;
Var
  i: Integer;
Begin
  For i := 0 To List.Count-1 Do
    List.Objects[i].Free;

  List.Clear;
End;


Procedure TWebFavoritesList.Clear;
Begin

  If Not Iterating Then CritSect.Enter;
  Try
    ClearPrim;
  Finally
    If Not Iterating Then CritSect.Leave;
  End;
End;


Function TWebFavoritesList.BuildIndexValue( Const Filename: String; Const Long: Boolean = False ): String;
Begin
  If Long Then
    Result := StringReplace( Filename, RootDirectory, '', [rfReplaceAll, rfIgnoreCase] )
   Else
    Result := ExtractRelativePath( ExtractShortPathName( RootDirectory ),
                                   ExtractShortPathName( Filename ) );
End;


Function TWebFavoritesList.IndexValueToFullFileName( Const Index: String ): String;
Begin
  Result := RootDirectory + Index;
End;

Procedure TWebFavoritesList.AddPrim( WebFav: TWebFavorite );
Begin
  List.AddObject( BuildIndexValue(WebFav.FileName), WebFav );
End;


Procedure TWebFavoritesList.Add( WebFav: TWebFavorite );
Begin
  If Not Iterating Then CritSect.Enter;
  Try
    AddPrim( WebFav );
  Finally
    If Not Iterating Then CritSect.Leave;
  End;
End;


Procedure TWebFavoritesList.DeletePrim( WebFav: TWebFavorite );
Var
  i: Integer;
Begin
  i := List.IndexOfObject( WebFav );
  If i < 0 Then Exit;

  WebFav.Free;
  List.Delete( i );
End;


Procedure TWebFavoritesList.Delete( WebFav: TWebFavorite );
Begin
  If Not Iterating Then CritSect.Enter;
  Try
    DeletePrim( WebFav );
  Finally
    If Not Iterating Then CritSect.Leave;
  End;
End;


Procedure TWebFavoritesList.DeletePrim( Const Filename: String );
Var
  i: Integer;
  WebFav: TWebFavorite;
Begin
  i := List.IndexOf( BuildIndexValue( Filename ) );
  If i < 0 Then Exit;

  WebFav := List.Objects[i] As TWebFavorite;
  WebFav.Free;
  List.Delete( i );
End;


Procedure TWebFavoritesList.Delete( Const Filename: String );
Begin
  If Not Iterating Then CritSect.Enter;
  Try
    DeletePrim( Filename );
  Finally
    If Not Iterating Then CritSect.Leave;
  End;
End;


Function TWebFavoritesList.ItemPrim( Const Filename: String ): TWebFavorite;
Var
  i: Integer;
Begin
  Result := Nil;

  i := List.IndexOf( BuildIndexValue( Filename ) );
  If i < 0 Then Exit;

  Result := List.Objects[i] As TWebFavorite;
End;


Function TWebFavoritesList.Item( Const Filename: String ): TWebFavorite;
Begin
  If Not Iterating Then CritSect.Enter;
  Try
    Result := Item( Filename );
  Finally
    If Not Iterating Then CritSect.Leave;
  End;
End;


Function TWebFavoritesList.ItemPrim( Index: Integer ): TWebFavorite;
Begin
  Result := List.Objects[ Index ] As TWebFavorite;
End;


Function TWebFavoritesList.Item( Index: Integer ): TWebFavorite;
Begin
  If Not Iterating Then CritSect.Enter;
  Try
    Result := ItemPrim( Index );
  Finally
    If Not Iterating Then CritSect.Leave;
  End;
End;


Function TWebFavoritesList.CountPrim: Integer;
Begin
  Result := List.Count;
End;


Function TWebFavoritesList.Count: Integer;
Begin
  If Not Iterating Then CritSect.Enter;
  Try
    Result := CountPrim;
  Finally
    If Not Iterating Then CritSect.Leave;
  End;
End;


Procedure TWebFavoritesList.BeginIteration;
Begin
  CritSect.Enter;
  Iterating := True;
End;


Procedure TWebFavoritesList.EndIteration;
Begin
  CritSect.Leave;
  Iterating := False;
End;


Procedure TWebFavoritesList.PopulateFavoritesToTree( TreeView: TTreeView; ARootNode: TTreeNode;
                                                     FavoriteImageIndex, FolderImageIndex: Integer  );

  Procedure FillFavoritesTree( RootNode: TTreeNode; Const RootDirectory: String );
  Var
    SR: TSearchRec;
    Node: TTreeNode;
    WebFav: TWebFavorite;
  Begin
    If 0 = FindFirst( IncludeTrailingPathDelimiter( RootDirectory ) + '*.*', faAnyFile, SR ) Then
      Try
        Repeat
          If SR.Attr And faDirectory = faDirectory Then Begin
            If (SR.Name <> '.') And (SR.Name <> '..') Then Begin

              WebFav := TWebFavorite.Create;
              WebFav.FavType := WFT_Folder;
              WebFav.Description := ExtractFileName( SR.Name );
              WebFav.URL := '';
              WebFav.FileName := IncludeTrailingPathDelimiter( RootDirectory ) + SR.Name;
              WebFav.Expires := Now;
              WebFav.LastModified := Now;
              WebFav.LastChecked := EncodeDate( 1970, 1, 1 );;
              WebFav.UpdateInterval := 0;
              Add( WebFav );

              Node := TreeView.Items.AddChild( RootNode, WebFav.Description );
              Node.Data := WebFav;
              Node.ImageIndex := FolderImageIndex;
              Node.SelectedIndex := FolderImageIndex;
              FillFavoritesTree( Node, WebFav.FileName );
            End;
          End
           Else Begin
             WebFav := TWebFavorite.Create;
             WebFav.FavType := WFT_Favorite;
             WebFav.Description := ChangeFileExt( SR.Name, '' );
             WebFav.FileName := IncludeTrailingPathDelimiter( RootDirectory ) + SR.Name;
             WebFav.URL := ResolveUrlIntShCut( WebFav.FileName );
             WebFav.Expires := FileDateToDateTime( FileAge( WebFav.FileName ) );
             WebFav.LastModified := FileDateToDateTime( FileAge( WebFav.FileName ) );
             WebFav.LastChecked := EncodeDate( 1970, 1, 1 );
             WebFav.UpdateInterval := 0;
             Add( WebFav );

             Node := TreeView.Items.AddChild( RootNode, WebFav.Description );
             Node.Data := WebFav;
             Node.ImageIndex := FavoriteImageIndex;
             Node.SelectedIndex := FavoriteImageIndex;
           End;

        Until 0 <> FindNext( SR );
      Finally
        FindClose( SR );
      End;
  End;

Begin
  If ARootNode = Nil Then
    Exit;

  CritSect.Enter;
  Try
    Clear;
    ARootNode.DeleteChildren;
    FillFavoritesTree( ARootNode, GetFavoritesDirectory );
    ARootNode.AlphaSort( True );
  Finally
    CritSect.Leave;
  End;
End;


Procedure TWebFavoritesList.CheckAllFavorites( ARootNode: TTreeNode );
Var
  i, j: Integer;
  WebFav: TWebFavorite;
  HTTP: TIdHTTP;
  //Auth: TIdAuthentication;
  //AuthManager: TIdAuthenticationManager;
  fModified: Boolean;
  TmpDate: TDateTime;
Begin
  Try
    Self.BeginIteration;
    Try
      For i := 0 To Self.Count-1 Do Begin
        WebFav := Self.Item( i );
        If WebFav = Nil Then Continue;
        If WebFav.FavType = WFT_Folder Then Continue;

        Try
          HTTP := TIdHTTP.Create( Nil );
          Try
            HTTP.Request.LastModified := WebFav.LastModified;
            HTTP.Head( WebFav.URL );
            If (HTTP.ResponseCode = 200) Then Begin
              WebFav.Expires := StrInternetToDateTime( HTTP.Response.RawHeaders.Values[ 'Expires' ] );
              TmpDate := StrInternetToDateTime( HTTP.Response.RawHeaders.Values[ 'Date' ] );
              fModified := WebFav.LastModified <> TmpDate;
              If fModified Then
                WebFav.LastModified := TmpDate;
            End
             Else fModified := True;
          Finally
            HTTP.Free;
          End;
        Except
          fModified := False;
        End;


        If fModified Then Begin
          For j := 0 To ARootNode.Count-1 Do
            If TWebFavorite(ARootNode.Item[j].Data) = WebFav Then Begin
              ARootNode.Item[j].Text := '**' + ARootNode.Item[j].Text;
              InvalidateTreeNode( ARootNode.TreeView, ARootNode.Item[j] );
            End;
        End;

      End;
    Finally
      Self.EndIteration;
    End;
  Except
  End;
End;


Procedure TWebFavoritesList.SaveToFile( OwnerForm: TForm );
Var
  DomImplementation: TDomImplementation;
  DomDoc: tdomdocument;
  Element: TdomElement;
  cac: TDomElement;
  j1: Integer;
  WebFav: TWebFavorite;
begin
  BeginIteration;
  Try
    Try
      DomImplementation := Nil;
      DomDoc := Nil;
      Try
        DomImplementation := TDomImplementation.Create( Nil );
        DomDoc := TdomDocument.Create( DomImplementation, Nil );
        DomDoc.encoding := 'UTF-8';

        Element := DomDoc.createElement( 'webfavorites' );
        DomDoc.appendChild( Element );

        for j1 := 0 to Self.Count - 1 do Begin
          WebFav := Self.Item( j1 );

          cac := DomDoc.createElement( 'item' );
          Element.appendChild(cac);

          If WebFav.FavType = WFT_Folder Then
            AddChildToDomDoc( DomDoc, cac, 'type', 'folder' )
           Else
            AddChildToDomDoc( DomDoc, cac, 'type', 'favorite' );

          AddChildToDomDoc( DomDoc, cac, 'description', WebFav.Description );
          AddChildToDomDoc( DomDoc, cac, 'url', WebFav.URL );
          AddChildToDomDoc( DomDoc, cac, 'filename', BuildIndexValue( WebFav.FileName, True ) );
          AddChildToDomDoc( DomDoc, cac, 'lastmodified', DateTimeToInternetStr(WebFav.LastModified) );
          AddChildToDomDoc( DomDoc, cac, 'expires', DateTimeToInternetStr(WebFav.Expires) );
          AddChildToDomDoc( DomDoc, cac, 'interval', IntToStr(WebFav.UpdateInterval) );
        end; //for j1

        SaveDomDocumentToFile( DomDoc, gProperties.DataDirectory + '\WebFavorites.xml' );

      Finally
        DomDoc.Free;
        DomImplementation.Free;
      End;
    Except
      On E: Exception Do
        Windows.MessageBox( OwnerForm.Handle, 'Error', PChar('Error saving WebFavorites.'#13#10#13#10 + E.message), MB_ICONHAND OR MB_OK );
    End;
  Finally
    EndIteration;
  End;
End;


Procedure TWebFavoritesList.LoadFromFile( OwnerForm: TForm );
var
  FileName: String;
  domdoc: tdomdocument;
  DomImplementation: tdomimplementation;
  XmlToDomParser: txmltodomparser;
  tase1, tase2: TdomNode;
  fAdvance: Boolean;
  WebFav: TWebFavorite;
  fAddIt: Boolean;
begin

  CritSect.Enter;
  Try
    Try
      ClearPrim;

      FileName := gProperties.DataDirectory + '\WebFavorites.xml';

      if not SysUtils.FileExists( FileName ) then
        exit;

      DomImplementation := Nil;
      XmlToDomParser := Nil;
      Try
         DomImplementation := TDomImplementation.Create(nil);
         XmlToDomParser := TXmlToDomParser.Create(nil);
         Xmltodomparser.DOMImpl := DomImplementation;

         domdoc := LoadFileIntoXDom( FileName, XmlToDomParser );

         tase1 := domdoc.documentElement.firstChild;

         while tase1<>nil do
         begin

          WebFav := Nil;
          fAdvance := True;
          fAddIt := False;
          tase2 := tase1.firstChild;
          while tase2<>nil do
          begin
            if tase2.nodeType = ntElement_Node then
            begin
              if fAdvance Then Begin
                fAdvance := False;
                WebFav := TWebFavorite.Create;
                WebFav.LastChecked := EncodeDate( 1970, 1, 1 );
                fAddIt := True;
              End;

                   if tase2.NodeName = 'type'    then Begin
                                                        If 0 = CompareText(tase2.textContent, 'Folder') Then
                                                          WebFav.FavType := WFT_Folder
                                                         Else
                                                          WebFav.FavType := WFT_Favorite;
                                                      End
              else if tase2.NodeName = 'description'  then WebFav.Description := tase2.textContent
              else if tase2.NodeName = 'url'          then WebFav.URL := tase2.textContent
              else if tase2.NodeName = 'filename'     then Webfav.FileName := IndexValueToFullFileName( tase2.textContent )
              else if tase2.NodeName = 'expires'      then WebFav.Expires := StrInternetToDateTime( tase2.textContent )
              else if tase2.NodeName = 'lastmodified' then WebFav.LastModified := StrInternetToDateTime( tase2.textContent )
              else if tase2.NodeName = 'interval'     then WebFav.UpdateInterval := StrToIntDef( tase2.textContent, 0 );
            end;

            tase2 := tase2.nextsibling;
          end;
          tase1 := tase1.NextSibling;

          If fAddIt Then
            AddPrim( WebFav );
        end;
      Finally
        XmlToDomParser.Free;
        DomImplementation.Free;
      End;
    Except
      On E: Exception Do Begin
        Windows.MessageBox( OwnerForm.Handle, 'Error', PChar('Error loading WebFavorites.'#13#10#13#10 + E.Message), MB_ICONHAND OR MB_OK );
      End;
    End;
  Finally
    CritSect.Leave;
  End;
End;


end.
